from plone.dexterity.content import Container
from seantis.people.interfaces import IPerson


class PersonBase(Container):

    @property
    def memberships(self):
        return IPerson(self).memberships()

    @property
    def organizations(self):
        return IPerson(self).organizations()

    @property
    def organization_uuids(self):
        return IPerson(self).organization_uuids()



from plone import api


def run_import_step_from_profile(
    context, step, module='seantis.people', profile='default'
):
    setup = api.portal.get_tool('portal_setup')
    profile = 'profile-{}:{}'.format(module, profile)
    setup.runImportStepFromProfile(profile, step)


def install_custom_controlpanel(context):
    run_import_step_from_profile(context, 'controlpanel')


def upgrade_phz_type_info(context):
    run_import_step_from_profile(context, 'typeinfo', 'seantis.people', 'phz')



from datetime import date
from plone import api
from plone.uuid.interfaces import IUUID
from zope.interface import Invalid

from seantis.people.interfaces import IMembership, IPerson
from seantis.people.content import Membership
from seantis.people import tests


class TestMembership(tests.IntegrationTestCase):

    def test_membership_creation(self):
        with self.user('admin'):
            ms = api.content.create(
                id='test',
                type='seantis.people.membership',
                container=self.new_temporary_folder()
            )

        self.assertIs(type(ms.aq_base), Membership)

    def test_membership_range_validation(self):
        with self.user('admin'):
            ms = api.content.create(
                id='test',
                type='seantis.people.membership',
                container=self.new_temporary_folder()
            )

        ms.start, ms.end = None, None
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 1), None
        IMembership.validateInvariants(ms)

        ms.start, ms.end = None, date(2012, 1, 1)
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 1), date(2012, 1, 2)
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 1), date(2012, 1, 1)
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 2), date(2012, 1, 1)
        self.assertRaises(Invalid, IMembership.validateInvariants, ms)

    def test_membership_person_relation(self):
        person_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__]
        )

        with self.user('admin'):
            organization = self.new_temporary_folder('S.H.I.E.L.D')

            nick_fury = api.content.create(
                title='Nick Fury',
                type=person_type.id,
                container=self.new_temporary_folder()
            )

            tony_stark = api.content.create(
                title='Tony Stark',
                type=person_type.id,
                container=self.new_temporary_folder()
            )

            api.content.create(
                title='Director',
                type='seantis.people.membership',
                container=organization,
                person=nick_fury
            )

            api.content.create(
                title='Head',
                type='seantis.people.membership',
                container=organization,
                person=tony_stark
            )

        memberships = IPerson(nick_fury).memberships()

        org = IUUID(organization)

        self.assertEqual(len(memberships), 1)
        self.assertEqual(memberships.keys(), [org])
        self.assertEqual(len(memberships[org]), 1)
        self.assertEqual(memberships[org][0].Title, 'Director')

        with self.user('admin'):
            api.content.create(
                title='Leutenant',
                type='seantis.people.membership',
                container=organization,
                person=nick_fury
            )

        memberships = IPerson(nick_fury).memberships()

        self.assertEqual(len(memberships), 1)
        self.assertEqual(memberships.keys(), [org])
        self.assertEqual(len(memberships[org]), 2)
        self.assertEqual(memberships[org][0].Title, 'Director')
        self.assertEqual(memberships[org][1].Title, 'Leutenant')



""" Provides a fast renderer for schema fields. Since this renderer is used
on lists and tables and therefore called many times in a request the following
options were not chosen:

z3c.form display mode rendering - too complicated, hard to customize and slow
zpt templates - quite a bit slower because of features we don't need

The snippets rendered here so simple that string.Template can be used which
is by far the fastest way of doing templates in python.

"""

import string

from zope.schema import getFields, Text, Tuple, List, Set, FrozenSet
from plone.namedfile.field import NamedBlobImage, NamedImage
from plone.app.textfield import RichText
from plone.app.uuid.utils import uuidToCatalogBrain
from Products.ZCatalog.interfaces import ICatalogBrain

from seantis.plonetools.schemafields import Email, Website
from seantis.plonetools import tools

from seantis.people.utils import UUIDList


class EmailFieldRenderer(object):

    template = string.Template(u'<a href="mailto:${mail}">${mail}</a>')

    def __call__(self, context, field):
        mail = getattr(context, field)
        if mail:
            return self.template.substitute(mail=mail)
        else:
            return u''


class WebsiteFieldRenderer(object):

    template = string.Template(u'<a href="${url}" target="_blank">${url}</a>')

    def __call__(self, context, field):
        url = getattr(context, field)
        if url:
            return self.template.substitute(url=url)
        else:
            return u''


class TextRenderer(object):

    def __call__(self, context, field):
        return '<br />'.join(getattr(context, field, u'').splitlines())


class RichTextRenderer(object):

    def __call__(self, context, field):
        return getattr(context, field).output


class ListRenderer(object):

    def __call__(self, context, field):
        return u', '.join(getattr(context, field, tuple()))


class UUIDListRenderer(object):

    template = string.Template(u'<a href="${url}">${title}</a>')

    def __call__(self, context, field):
        uuids = getattr(context, field, None)

        if not uuids:
            return u''

        unicode_sortkey = tools.unicode_collate_sortkey()

        brains = (uuidToCatalogBrain(uid) for uid in uuids)
        items = sorted(
            ((b.getURL(), b.Title) for b in brains),
            key=lambda i: unicode_sortkey(i[1])
        )

        return ', '.join(
            self.template.substitute(url=url, title=title.decode('utf-8'))
            for url, title in items
        )


class ImageRenderer(object):

    template = string.Template(u'<img src="${url}" />')

    def __call__(self, context, field):
        img = getattr(context, field)
        if img:
            if ICatalogBrain.providedBy(context):
                baseurl = context.getURL()
            else:
                baseurl = context.absolute_url()

            url = '/'.join((baseurl, '@@images', field, 'thumb'))

            return self.template.substitute(url=url)
        else:
            return u''


# This is not the best way to match objects to classes, but it sure is the
# fastest. Checking through isinstance would require going through a list.
# Since that can easily happen more than a thousand times in a request it's
# better to be fast than to be right.
renderers = {
    Email: EmailFieldRenderer(),
    Website: WebsiteFieldRenderer(),
    NamedBlobImage: ImageRenderer(),
    NamedImage: ImageRenderer(),
    Text: TextRenderer(),
    RichText: RichTextRenderer(),
    Tuple: ListRenderer(),
    List: ListRenderer(),
    Set: ListRenderer(),
    FrozenSet: ListRenderer(),
    list: ListRenderer(),
    set: ListRenderer(),
    tuple: ListRenderer(),
    UUIDList: UUIDListRenderer(),
}


class Renderer(object):

    def __init__(self, schema, redirects=None):
        self.schema = schema
        self.fields = getFields(schema)
        self.redirects = redirects or {}

    def render(self, context, field):
        field = self.redirects.get(field, field)
        fieldtype = type(self.fields.get(field, getattr(context, field, None)))
        return renderers.get(fieldtype, getattr)(context, field)



from datetime import date, MAXYEAR

from plone import api
from zope.interface import implements

from seantis.plonetools import tools
from seantis.people.interfaces import IPerson
from seantis.people.behaviors.memberships import get_memberships
from seantis.people.utils import UUIDList

class Person(object):
    implements(IPerson)

    def __init__(self, context):
        self.context = context

    def memberships(self):
        return get_memberships(self.context)

    def organizations(self):
        """ Returns a list of organizations this person is active in. The list
        is sorted by uca collation. This is mainly used for the organizations
        metadata field of all person objects (to show in the list).

        """

        organizations = self.organization_uuids()

        catalog = api.portal.get_tool('portal_catalog')
        titles = (catalog(UID=uid)[0].Title for uid in organizations)

        return sorted(titles, key=tools.unicode_collate_sortkey())

    def organization_uuids(self):
        """ Returns the organizations as uuid references which are be rendered
        as links in the list view.

        """
        organizations = UUIDList()

        for uuid, memberships in self.memberships().items():
            if not self.active_memberships(memberships):
                continue

            organizations.append(uuid)

        return organizations

    def active_memberships(self, memberships):
        """ Goes through the given memberships and returns the active ones.
        The active memberships are the ones without end date. Or with an end
        date in the future.

        The result is sorted by date. Memberships without end come before
        memberships with end.
        """

        def sortkey(membership):
            if membership.end is not None:
                return membership.end
            else:
                return date(MAXYEAR, 12, 31)

        today = date.today()
        active = [m for m in memberships if m.end is None or m.end >= today]
        return sorted(active, key=sortkey)

    def current_role(self, memberships):
        """ Goes through the given memberships and returns the current role.
        The current role is the latest active membership's role.

        """
        active = self.active_memberships(memberships)
        if active:
            return active[0].role
        else:
            return u''



from five import grok

from zope.component import getAdapters
from zope.interface import Interface
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent
from plone import api
from plone.uuid.interfaces import IUUID
from plone.indexer.decorator import indexer

from seantis.plonetools import tools
from seantis.people.interfaces import IMembershipSource, IMembership


@indexer(IMembership)
def membership_person(membership, **kw):
    """ Stores the person on each membership object for faster lookups. """
    return IUUID(membership.person)


def on_membership_changed(event):
    """ Listens to IMembershipChanged events relaying them to the Catalog
    for reindexing of the person whose membership changed.

    """
    notify(ObjectModifiedEvent(event.person))


def get_memberships(person=None):
    context = api.portal.get()

    organizations = {}
    for name, source in getAdapters((context, ), IMembershipSource):
        for organization, memberships in source.memberships(person).items():

            if organization not in organizations:
                organizations[organization] = []

            organizations[organization].extend(memberships)

        organizations.update(source.memberships(person))

    return organizations


class ZodbMembershipSource(grok.Adapter):

    grok.name('zodb-membership-source')
    grok.provides(IMembershipSource)
    grok.context(Interface)

    def memberships(self, person=None):

        query = {'portal_type': 'seantis.people.membership'}

        if person is not None:
            query['membership_person'] = IUUID(person)

        memberships = api.portal.get_tool(name='portal_catalog')(**query)

        result = {}

        for membership in memberships:
            organization = tools.get_parent(membership).UID
            result.setdefault(organization, []).append(membership)

        return result
